# limitations under the License.
""" Utilities for the standard evaluator and its ability to save and load state"""

def unique_names(var):
    names = []
    for my_var in var:
        names.append(my_var.name)
    if len(names) != len(set(names)):
        dup = {x for x in names if names.count(x) > 1}
        raise ValueError(
            f"There is at least one variable / response name that is used multiple times. Names are {dup}"
        )



import openmdao.api as om
from openmdao.test_suite.scripts.circuit_analysis import Resistor, Diode, Node


class Circuit(om.Group):
    def setup(self):
        self.add_subsystem(
            "n1", Node(n_in=1, n_out=2), promotes_inputs=[("I_in:0", "I_in")]
        )
        self.add_subsystem("n2", Node())  # leaving defaults

        self.add_subsystem("R1", Resistor(R=100.0), promotes_inputs=[("V_out", "Vg")])
        self.add_subsystem("R2", Resistor(R=10000.0))
        self.add_subsystem("D1", Diode(), promotes_inputs=[("V_out", "Vg")])

        self.connect("n1.V", ["R1.V_in", "R2.V_in"])
        self.connect("R1.I", "n1.I_out:0")
        self.connect("R2.I", "n1.I_out:1")

        self.connect("n2.V", ["R2.V_out", "D1.V_in"])
        self.connect("R2.I", "n2.I_in:0")
        self.connect("D1.I", "n2.I_out:0")

        self.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        self.nonlinear_solver.options["iprint"] = 2
        self.nonlinear_solver.options["maxiter"] = 20
        self.linear_solver = om.DirectSolver()

    def create_model(self):
        # Setup problem and model
        self.p = om.Problem()
        self.model = self.p.model
        #
        self.model.set_input_defaults("ground.V", 0.0, units="V")
        self.model.set_input_defaults("source.I", 0.1, units="A")
        self.model.add_subsystem(
            "circuit",
            Circuit(),
            promotes_inputs=[("Vg", "ground.V"), ("I_in", "source.I")],
        )
        #
        self.model.add_design_var("ground.V")
        self.model.add_design_var("source.I")
        self.model.add_objective("circuit.D1.I")

        self.p.setup()
        self.p.check_config(checks=["unconnected_inputs"], out_file=None)
        #
        # set some initial guesses
        self.p["circuit.n1.V"] = 10.0
        self.p["circuit.n2.V"] = 1.0



import copy
import pytest
import pandas as pd
import numpy as np
from pydantic import ValidationError
from standard_evaluator import FloatVariable, IntVariable, ArrayVariable, CategoricalVariable
from standard_evaluator import EvaluatorInfo, GroupInfo, JoinedInfo,OptProblem, MAXINT

def test_evaluator_info():
    with pytest.raises(ValueError):
        # Check that if if you try to name two inputs the same it throws an error
        EvaluatorInfo(name="bla", inputs=[{"name": "x0"}, {"name": "x0"}])

    with pytest.raises(ValueError):
        # Check that if if you try to name two outputs the same it throws an error
        EvaluatorInfo(name="bla", outputs=[{"name": "x0"}, {"name": "x0"}])
    my_eval = EvaluatorInfo(
        name="dummy",
        inputs=[
            ArrayVariable(name="x1", default=[[299.3, 243.5]]),
            FloatVariable(name="p2", class_type="float"),
        ],
        outputs=[FloatVariable(name="y2", class_type="float")],
    )
    expected = {
        "name": "dummy",
        "class_type": "EvaluatorInfo",
        "inputs": [
            {
                "name": "x1",
                "default": np.array([[299.3, 243.5]]),
                "bounds": None,
                "shift": None,
                "scale": None,
                "units": None,
                "description": "",
                "class_type": "floatarray",
                "shape": (1, 2),
                "options": {},
            },
            {
                "name": "p2",
                "default": None,
                "bounds": (-np.inf, np.inf),
                "shift": 0.0,
                "scale": 1.0,
                "units": None,
                "description": "",
                "class_type": "float",
                "options": {},
            },
        ],
        "outputs": [
            {
                "name": "y2",
                "default": None,
                "bounds": (-np.inf, np.inf),
                "shift": 0.0,
                "scale": 1.0,
                "units": None,
                "description": "",
                "class_type": "float",
                "options": {},
            }
        ],
        "description": None,
        "cite": None,
        "tool": None,
        "evaluator_identifier": None,
        "version": None,
        "component_type": None,
        "options": {},
    }
    np.testing.assert_equal(expected, my_eval.model_dump())

def test_group():
    x1 = ArrayVariable(name="x1", default=[[299.3, 243.5]])
    p2 = FloatVariable(name="p2")
    q1 = FloatVariable(name="q1")
    y2 = FloatVariable(name="y2")
    y4 = FloatVariable(name="y4")
    my_eval1 = EvaluatorInfo(
            name="Part1",
            inputs=[x1, p2, q1],
            outputs=[y2],
        )
    my_eval2 = EvaluatorInfo(
            name="Part2",
            inputs=[x1, p2, y2],
            outputs=[y4],
        )
    my_group = GroupInfo(name="MyGroup",
            inputs=[x1, p2, q1],
            outputs=[y4],
            components={my_eval1.name: my_eval1, my_eval2.name: my_eval2},
            component_order=[my_eval1.name, my_eval2.name],
            linkage=[("Part1.x1", "x1"), ("Part1.p2", "p2"), ("Part1.q1", "q1"),
                    ("Part1.y2", "Part2.y2"), ("Part2.x1", "x1"), ("Part2.p2", "p2"),
                    ("Part2.y4", "y4")])
    expected = {'name': 'MyGroup',
 'class_type': 'GroupInfo',
 'inputs': [{'name': 'x1',
   'default': np.array([[299.3, 243.5]]),
   'bounds': None,
   'shift': None,
   'scale': None,
   'units': None,
   'description': '',
   'options': {},
   'class_type': 'floatarray',
   'shape': (1, 2)},
  {'name': 'p2',
   'default': None,
   'bounds': (-np.inf, np.inf),
   'shift': 0.0,
   'scale': 1.0,
   'units': None,
   'description': '',
   'options': {},
   'class_type': 'float'},
  {'name': 'q1',
   'default': None,
   'bounds': (-np.inf, np.inf),
   'shift': 0.0,
   'scale': 1.0,
   'units': None,
   'description': '',
   'options': {},
   'class_type': 'float'}],
 'outputs': [{'name': 'y4',
   'default': None,
   'bounds': (-np.inf, np.inf),
   'shift': 0.0,
   'scale': 1.0,
   'units': None,
   'description': '',
   'options': {},
   'class_type': 'float'}],
 'description': None,
 'cite': None,
 'tool': None,
 'evaluator_identifier': None,
 'version': None,
 'component_type': None,
 'options': {},
 'component_order': ['Part1', 'Part2'],
 'components': {'Part1': {'name': 'Part1',
   'class_type': 'EvaluatorInfo',
   'inputs': [{'name': 'x1',
     'default': np.array([[299.3, 243.5]]),
     'bounds': None,
     'shift': None,
     'scale': None,
     'units': None,
     'description': '',
     'options': {},
     'class_type': 'floatarray',
     'shape': (1, 2)},
    {'name': 'p2',
     'default': None,
     'bounds': (-np.inf, np.inf),
     'shift': 0.0,
     'scale': 1.0,
     'units': None,
     'description': '',
     'options': {},
     'class_type': 'float'},
    {'name': 'q1',
     'default': None,
     'bounds': (-np.inf, np.inf),
     'shift': 0.0,
     'scale': 1.0,
     'units': None,
     'description': '',
     'options': {},
     'class_type': 'float'}],
   'outputs': [{'name': 'y2',
     'default': None,
     'bounds': (-np.inf, np.inf),
     'shift': 0.0,
     'scale': 1.0,
     'units': None,
     'description': '',
     'options': {},
     'class_type': 'float'}],
   'description': None,
   'cite': None,
   'tool': None,
   'evaluator_identifier': None,
   'version': None,
   'component_type': None,
   'options': {}},
  'Part2': {'name': 'Part2',
   'class_type': 'EvaluatorInfo',
   'inputs': [{'name': 'x1',
     'default': np.array([[299.3, 243.5]]),
     'bounds': None,
     'shift': None,
     'scale': None,
     'units': None,
     'description': '',
     'options': {},
     'class_type': 'floatarray',
     'shape': (1, 2)},
    {'name': 'p2',
     'default': None,
     'bounds': (-np.inf, np.inf),
     'shift': 0.0,
     'scale': 1.0,
     'units': None,
     'description': '',
     'options': {},
     'class_type': 'float'},
    {'name': 'y2',
     'default': None,
     'bounds': (-np.inf, np.inf),
     'shift': 0.0,
     'scale': 1.0,
     'units': None,
     'description': '',
     'options': {},
     'class_type': 'float'}],
   'outputs': [{'name': 'y4',
     'default': None,
     'bounds': (-np.inf, np.inf),
     'shift': 0.0,
     'scale': 1.0,
     'units': None,
     'description': '',
     'options': {},
     'class_type': 'float'}],
   'description': None,
   'cite': None,
   'tool': None,
   'evaluator_identifier': None,
   'version': None,
   'component_type': None,
   'options': {}}},
 'linkage': [('Part1.x1', 'x1'),
  ('Part1.p2', 'p2'),
  ('Part1.q1', 'q1'),
  ('Part1.y2', 'Part2.y2'),
  ('Part2.x1', 'x1'),
  ('Part2.p2', 'p2'),
  ('Part2.y4', 'y4')],
  'promotions': {},
  }
    np.testing.assert_equal(expected, my_group.model_dump())
  
def test_evaluator_info_default():
    lower_bound = np.array([[1, 2], [-np.inf, 4], [5, 4]])
    upper_bound = np.array([[3, 4], [6, np.inf], [np.inf, 8]])
    expected_default = np.array([[2., 3.], [6., 4.], [5., 6.]])
    x1 = FloatVariable(name='x1', bounds=[2.,5])
    x2 = IntVariable(name='x2', bounds=[2,5])
    cat = CategoricalVariable(name="cat1", bounds=["x2", "f3", "f1>1"])
    array1 = ArrayVariable(name="array1", bounds=(lower_bound, upper_bound), shape=lower_bound.shape)
    y1 = FloatVariable(name='y1', bounds=[2.,5])
    y2 = IntVariable(name='y2', bounds=[2,5])
    my_prob = EvaluatorInfo(name="opt", inputs=[x1, x2, cat, array1], outputs=[y1, y2])
    my_prob.calculate_default()
    assert my_prob.inputs[0].default == 3.5
    assert my_prob.inputs[1].default == 3
    assert my_prob.inputs[2].default == "x2"
    np.testing.assert_equal(my_prob.inputs[3].default, expected_default)

    with pytest.raises(ValidationError):
        my_prob.set_defaults({'x1': 4.0, 'x2': 'sd', 'cat1': 'f3'})


import re
import importlib
import pprint
import json
import typing
from typing import List, Optional, Dict, Tuple, Union, Literal
import copy
import inspect
from enum import Enum
from collections.abc import Callable
from base64 import b64encode
import inspect
import pathlib

import json_numpy
from pydantic import BaseModel, Field, field_validator, model_validator
import numpy as np
from numpy import generic, ndarray
from numpy.lib.format import descr_to_dtype, dtype_to_descr
from numpydantic import NDArray, Shape

import openmdao
import openmdao.api as om
from openmdao.core.constants import _ReprClass
import dymos

import aviary
from aviary.variable_info.enums import ProblemType, EquationsOfMotion, LegacyCode
from aviary.utils.aviary_values import AviaryValues
from aviary.subsystems.propulsion.engine_deck import EngineDeck 

class AviaryEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, type):
            return {'__type__': str(obj)}
        elif isinstance(obj, set):
            return {'__set__': list(obj)}
        elif isinstance(obj, Enum):
            return {'__enum__': {'type': str(type(obj)),
                                 'value': repr(obj),
                                 'module': str(inspect.getmodule(obj))}}
        elif isinstance(obj, (ndarray, generic)):
            data = obj.data if obj.flags["C_CONTIGUOUS"] else obj.tobytes()
            return {
                "__numpy__": b64encode(data).decode(),
                "dtype": dtype_to_descr(obj.dtype),
                "shape": obj.shape,
            }
        elif isinstance(obj, tuple):
            return {'__tuple__': list(obj)}
        elif isinstance(obj, _ReprClass):
            return {'__repr_class__': str(obj)}
        elif isinstance(obj, Callable):
            # For a callable (function) we need to get the name and module so we can
            # load the function.
            func_name = str(obj).split(' ')[1]
            module_name = str(inspect.getmodule(obj)).split("'")[1]
            return {'__Callable__': {
                'name': func_name,
                'module': module_name,
            }}
        elif isinstance(obj, type(...)):
            return {'__Ellipsis__': 'Use an ellipsis'}
        elif isinstance(obj, pathlib.WindowsPath):
            # A Windows path can easily be translated into a string
            return {'__pathlib.WindowsPath__': str(obj)}
        elif isinstance(obj, AviaryValues):
            # An AviaryValues instance can be converted into a dictionary which
            # then gets unrolled recursively
            return {'__aviary_values__': dict(obj)}
        elif isinstance(obj, om.OptionsDictionary):
            # We want to make sure we can also handle descendants of the OptionsDictionary.
            # Therefore we are storing the actual type of the class
            return {'__openmdao.api.OptionsDictionary__':
                    {'type': str(type(obj)),
                     'value': obj.__getstate__()}
                }
        elif isinstance(obj, aviary.subsystems.aerodynamics.aerodynamics_builder.CoreAerodynamicsBuilder):
            return {'__CoreAerodynamicsBuilder__': 'Not implemented'}
        elif isinstance(obj, aviary.subsystems.propulsion.engine_deck.EngineDeck):
            return {'__EngineDeck__': {'type': str(type(obj)),
                                 'module': str(inspect.getmodule(obj)),
                                 'options': obj.options}}
        elif isinstance(obj, aviary.subsystems.propulsion.propulsion_builder.CorePropulsionBuilder):
            return {'__CorePropulsionBuilder__': 'Not implemented'}
        elif isinstance(obj, dymos.Radau):
            return {'__dymos.Radau__': 'Not implemented'}
        elif isinstance(obj, dymos.transcriptions.grid_data.GridData):
            return {'__dymos.transcriptions.grid_data.GridData__': 'Not implemented'}
        print(f'class type {type(obj)} from module {inspect.getmodule(obj)}')
        # Let the base class default method raise the TypeError
        return super().default(obj)


import copy

import openmdao.core.group

"""Code for generating model definition using N2 viewer framework."""
from operator import itemgetter

import networkx as nx

from openmdao.core.explicitcomponent import ExplicitComponent
from openmdao.core.indepvarcomp import IndepVarComp
from openmdao.core.group import Group
from openmdao.core.problem import Problem, _SetupStatus
from openmdao.core.implicitcomponent import ImplicitComponent
from openmdao.core.constants import _UNDEFINED
from openmdao.components.exec_comp import ExecComp
from openmdao.components.meta_model_structured_comp import MetaModelStructuredComp
from openmdao.components.meta_model_unstructured_comp import MetaModelUnStructuredComp
from openmdao.drivers.doe_driver import DOEDriver
from openmdao.recorders.case_reader import CaseReader

# from openmdao.utils.array_utils import convert_ndarray_to_support_nans_in_json
from openmdao.utils.class_util import overrides_method
from openmdao.utils.general_utils import default_noraise
from openmdao.utils.om_warnings import issue_warning
from openmdao import __version__ as openmdao_version

# from typing import Union

import json
import numpy as np
import json_numpy

json_numpy.patch()


def convert_nans_in_nested_list(val_as_list: list) -> None:
    """
    Given a list, possibly nested, replace any numpy.nan values with the string "nan".

    This is done since JSON does not handle nan. This code is used to pass variable values
    to the N2 diagram.

    The modifications to the list values are done in-place to avoid excessive copying of lists.

    Parameters
    ----------
    val_as_list : list
        List, possibly nested, whose nan elements need to be converted.
    """
    for i, val in enumerate(val_as_list):
        if isinstance(val, list):
            convert_nans_in_nested_list(val)
        else:
            if np.isnan(val):
                val_as_list[i] = "nan"
            elif np.isinf(val):
                val_as_list[i] = "infinity"
            else:
                val_as_list[i] = val


def convert_ndarray_to_support_nans_in_json(val: np.array) -> list:
    """
    Given numpy array of arbitrary dimensions, return the equivalent nested list with nan replaced.

    numpy.nan values are replaced with the string "nan".

    Parameters
    ----------
    val : ndarray
        Numpy array to be converted.

    Returns
    -------
    list
        The equivalent list (possibly nested) with any nan values replaced with the string "nan".
    """
    val = np.asarray(val)

    # do a quick check for any nans or infs and if not we can avoid the slow check
    nans = np.where(np.isnan(val))
    infs = np.where(np.isinf(val))
    if nans[0].size == 0 and infs[0].size == 0:
        return val.tolist()

    val_as_list = val.tolist()
    convert_nans_in_nested_list(val_as_list)
    return val_as_list


def _get_array_info(
    system: openmdao.core.group.System,
    vec: str,
    name: str,
    prom: str,
    var_dict: dict,
    from_src: bool = True,
) -> None:
    """

    Parameters
    ----------
    system : System
        The System at the root of the hierarchy
    vec: str
        Absolute name in the owning system's namespace.
    name: String
        Name of the value to get
    prom: str
        System component individual input or output
    var_dict: dict
        Dict of variables of prom
    from_src: bool
        If True, retrieve value of an input variable from its connected source.


    Returns
    -------
    """
    # if 'vec' is not None at this point, we can retrieve the value using vec._abs_get_val,
    # which is a faster call than system.get_val.
    ndarray_to_convert = (
        vec._abs_get_val(name, flat=False)
        if vec
        else system.get_val(prom, from_src=from_src)
    )

    var_dict["val"] = convert_ndarray_to_support_nans_in_json(ndarray_to_convert)

    # Find the minimum indices and value
    min_indices = np.unravel_index(
        np.nanargmin(ndarray_to_convert, axis=None), ndarray_to_convert.shape
    )
    var_dict["val_min_indices"] = min_indices
    var_dict["val_min"] = ndarray_to_convert[min_indices]

    # Find the maximum indices and value
    max_indices = np.unravel_index(
        np.nanargmax(ndarray_to_convert, axis=None), ndarray_to_convert.shape
    )
    var_dict["val_max_indices"] = max_indices
    var_dict["val_max"] = ndarray_to_convert[max_indices]


def _get_var_dict(
    system: openmdao.core.group.System,
    typ: str,
    name: str,
    is_implicit: bool,
    values: bool,
) -> dict:
    """
    Create dict from system model

    Parameters:
    ----------
    system: openmdao.core.group.System
        An openmdao System
    typ: str
        Variable type "input" or "output".
    name: str
        Node name
    is_implicit: bool
        Indicates if this is an implicit variable
    values: bool
        Indicates if values exist

    Returns:
    ----------
    var_dict: dict
        Variable dictionary from openmdao System
    """
    if name in system._var_abs2meta[typ]:
        meta = system._var_abs2meta[typ][name]
        prom = system._var_abs2prom[typ][name]
        val = np.asarray(meta["val"])
        is_dist = meta["distributed"]

        var_dict = {
            "name": prom,
            "type": typ,
            "dtype": type(val).__name__,
            "is_discrete": False,
            "distributed": is_dist,
            "shape": str(meta["shape"]),
            "desc": meta["desc"],
        }

        if typ == "output":
            var_dict["implicit"] = is_implicit
            vec = system._outputs
        else:  # input
            vec = None

        # if 'vec' is not None at this point, we can retrieve the value using vec._abs_get_val,
        # which is a faster call than system.get_val.

        if meta["units"] is None:
            var_dict["units"] = "None"
        else:
            var_dict["units"] = meta["units"]

        try:
            if values:
                # Get the current value
                _get_array_info(system, vec, name, prom, var_dict, from_src=True)
        except Exception as err:
            issue_warning(str(err))
    else:  # discrete
        meta = system._var_discrete[typ][name]
        val = meta["val"]
        var_dict = {
            "name": name,
            "type": typ,
            "dtype": type(val).__name__,
            "is_discrete": True,
        }
        if values:
            if isinstance(val, (int, str, list, dict, complex, np.ndarray)):
                var_dict["val"] = default_noraise(system.get_val(name))

    if "surrogate_name" in meta:
        var_dict["surrogate_name"] = meta["surrogate_name"]

    return var_dict


def _serialize_single_option(option: dict) -> object:
    #  -> Union[str, int, float, bool, list, tuple] :
    # tested
    """
    Return a json-safe equivalent of the option.

    The default_noraise function performs the datatype serialization, while this function takes
    care of attributes specific to options dicts.

    Parameters
    ----------
    option : object
        Option to be serialized.

    Returns
    -------
    object
       JSON-safe serialized object.
    """
    opt_keys = option.keys()
    if "recordable" not in opt_keys:
        return "Not Recordable"

    val = option["val"]

    if val is _UNDEFINED:
        return str(val)

    return default_noraise(val)


def _get_tree_dict(
    system: openmdao.core.group.System, values: bool = True, is_parallel: bool = False
) -> dict:
    """
    Get a dictionary representation of the system hierarchy.

    Parameters
    ----------
    system : System
        The System at the root of the hierarchy
    values : bool
        If True, include variable values. If False, all values will be None.
    is_parallel : bool
        If True, values can be remote and are not available.

    Returns
    ----------
    tree_dict: dict
        Returns tree dictionary of model
    """
    tree_dict = {
        "name": system.name if system.name else "root",
        "type": "subsystem" if system.name else "root",
        "class": ":".join((type(system).__module__, type(system).__qualname__)),
        "expressions": None,
        "nonlinear_solver": "",
        "nonlinear_solver_options": None,
        "linear_solver": "",
        "linear_solver_options": None,
    }
    is_implicit = False

    tree_dict["subsystem_type"] = "component"
    tree_dict["is_parallel"] = is_parallel
    if isinstance(system, ImplicitComponent):
        is_implicit = True
        tree_dict["component_type"] = "implicit"
        if overrides_method("solve_linear", system, ImplicitComponent):
            tree_dict["linear_solver"] = "solve_linear"
        elif system.linear_solver:
            tree_dict["linear_solver"] = system.linear_solver.SOLVER
            tree_dict["linear_solver_options"] = {
                k: _serialize_single_option(opt)
                for k, opt in system.linear_solver.options._dict.items()
            }

        if overrides_method("solve_nonlinear", system, ImplicitComponent):
            tree_dict["nonlinear_solver"] = "solve_nonlinear"
        elif system.nonlinear_solver:
            tree_dict["nonlinear_solver"] = system.nonlinear_solver.SOLVER
            tree_dict["nonlinear_solver_options"] = {
                k: _serialize_single_option(opt)
                for k, opt in system.nonlinear_solver.options._dict.items()
            }
    elif isinstance(system, ExecComp):
        tree_dict["component_type"] = "exec"
        tree_dict["expressions"] = system._exprs
    elif isinstance(system, (MetaModelStructuredComp, MetaModelUnStructuredComp)):
        tree_dict["component_type"] = "metamodel"
    elif isinstance(system, IndepVarComp):
        tree_dict["component_type"] = "indep"
    elif isinstance(system, ExplicitComponent):
        tree_dict["component_type"] = "explicit"
    else:
        tree_dict["component_type"] = None

    children = []
    for typ in ["input", "output"]:
        for abs_name in system._var_abs2meta[typ]:
            children.append(_get_var_dict(system, typ, abs_name, is_implicit, values))

        for prom_name in system._var_discrete[typ]:
            children.append(_get_var_dict(system, typ, prom_name, is_implicit, values))

    tree_dict["children"] = children

    options = {}
    slv = {"linear_solver", "nonlinear_solver"}
    for k, opt in system.options._dict.items():
        if k in slv:
            # need to handle solver option separately because it can be a class, instance or None
            try:
                val = opt["val"]
            except KeyError:
                val = opt["value"]

            try:
                options[k] = val.SOLVER
            except AttributeError:
                options[k] = val
        else:
            options[k] = _serialize_single_option(opt)

    tree_dict["options"] = options

    return tree_dict


def _get_declare_partials(system: openmdao.core.group.System) -> list:
    # tested
    """
    Get a list of the declared partials.

    Parameters
    ----------
    system : <System>
        A System in the model.

    Returns
    -------
    list
        A list containing all the declared partials (strings in the form "of > wrt" )
        beginning from the given system on down.
    """
    declare_partials_list = []
    for of, wrt in system._declared_partials_iter():
        if of != wrt:
            declare_partials_list.append(f"{of} > {wrt}")

    return declare_partials_list


def _get_viewer_data(
    data_source: openmdao.core.problem.Problem,
    values: bool = _UNDEFINED,
    case_id: int or None = None,
):
    """
    Get the data needed by the N2 viewer as a dictionary.

    Parameters
    ----------
    data_source : <Problem> or <Group> or str
        A Problem or Group or case recorder filename containing the model or model data.
        If the case recorder file from a parallel run has separate metadata, the
        filenames can be specified with a comma, e.g.: case.sql_0,case.sql_meta
    values : bool or _UNDEFINED
        If True, include variable values. If False, all values will be None.
        If unspecified, this behaves as if set to True unless the data source is a Problem or
        model for which setup is not complete, in which case it behaves as if set to False.
    case_id : int or str or None
        Case name or index of case in SQL file.

    Returns
    -------
    dict
        A dictionary containing information about the model for use by the viewer.
    """
    if isinstance(data_source, Problem):
        root_group = data_source.model

        if not isinstance(root_group, Group):
            # this function only makes sense when the model is a Group
            msg = (
                f"The model is of type {root_group.__class__.__name__}, "
                "viewer data is only available if the model is a Group."
            )
            raise TypeError(msg)

        driver = data_source.driver
        driver_name = driver.__class__.__name__
        driver_type = "doe" if isinstance(driver, DOEDriver) else "optimization"

        driver_options = {
            key: _serialize_single_option(driver.options._dict[key])
            for key in driver.options
        }

        if driver_type == "optimization" and hasattr(driver, "opt_settings"):
            driver_opt_settings = driver.opt_settings
        else:
            driver_opt_settings = None

        # set default behavior for values flag
        if values is _UNDEFINED:
            values = (
                data_source._metadata is not None
                and data_source._metadata["setup_status"]
                >= _SetupStatus.POST_FINAL_SETUP
            )

    elif isinstance(data_source, Group):
        if not data_source.pathname:  # root group
            root_group = data_source
            driver_name = None
            driver_type = None
            driver_options = None
            driver_opt_settings = None
        else:
            # this function only makes sense when it is at the root
            msg = (
                f"Viewer data is not available for sub-Group '{data_source.pathname}'."
            )
            raise TypeError(msg)

        # set default behavior for values flag
        if values is _UNDEFINED:
            values = (
                data_source._problem_meta is not None
                and data_source._problem_meta["setup_status"]
                >= _SetupStatus.POST_FINAL_SETUP
            )

    elif isinstance(data_source, str):
        if "," in data_source:
            filenames = data_source.split(",")
            cr = CaseReader(filenames[0], metadata_filename=filenames[1])
        else:
            cr = CaseReader(data_source)

        data_dict = cr.problem_metadata

        # set default behavior for values flag
        if values is _UNDEFINED:
            values = True

        def set_values(children, stack, case_):
            """
            Set variable values in model tree from the specified Case.

            If case is None, set all values to None.
            """
            for child in children:
                # if 'val' in child
                if child["type"] == "subsystem":
                    stack.append(child["name"])
                    set_values(child["children"], stack, case_)
                    stack.pop()
                elif child["type"] == "input":
                    if case_ is None:
                        child.pop("val")
                        for key in [
                            "val_min",
                            "val_max",
                            "val_min_indices",
                            "val_max_indices",
                        ]:
                            del child[key]
                    elif case_.inputs is None:
                        child["val"] = "N/A"
                    else:
                        path = (
                            child["name"]
                            if not stack
                            else ".".join(stack + [child["name"]])
                        )
                        child["val"] = case_.inputs[path]
                elif child["type"] == "output":
                    if case_ is None:
                        child.pop("val")
                        for key in [
                            "val_min",
                            "val_max",
                            "val_min_indices",
                            "val_max_indices",
                        ]:
                            del child[key]
                    elif case_.outputs is None:
                        child["val"] = "N/A"
                    else:
                        path = (
                            child["name"]
                            if not stack
                            else ".".join(stack + [child["name"]])
                        )
                        try:
                            child["val"] = case_.outputs[path]
                        except KeyError:
                            child["val"] = "N/A"

        if values is False:
            set_values(data_dict["tree"]["children"], [], None)
        elif case_id is not None:
            case = cr.get_case(case_id)
            print(f"Using source: {case.source}\nCase: {case.name}")
            set_values(data_dict["tree"]["children"], [], case)

        # Delete the variables key since it's not used in N2
        if "variables" in data_dict:
            del data_dict["variables"]

        # Older recordings might not have this.
        if "md5_hash" not in data_dict:
            data_dict["md5_hash"] = None

        return data_dict

    else:
        raise TypeError(
            f"Viewer data is not available for '{data_source}'."
            "The source must be a Problem, model or the filename of a recording."
        )

    data_dict = {
        "tree": _get_tree_dict(root_group, values=values),
        "md5_hash": root_group._generate_md5_hash(),
    }

    connections_list = []

    grp = root_group.compute_sys_graph(comps_only=True)

    scc = nx.strongly_connected_components(grp)

    strongdict = {}
    sys_idx_names = []

    for i, strong_comp in enumerate(scc):
        for c in strong_comp:
            strongdict[c] = i  # associate each comp with a strongly connected component

        if len(strong_comp) > 1:
            # these IDs are only used when back edges are present
            for name in strong_comp:
                sys_idx_names.append(name)

    sys_idx = (
        {}
    )  # map of pathnames to index of pathname in list (systems in cycles only)

    comp_orders = {
        name: i for i, name in enumerate(root_group._ordered_comp_name_iter())
    }
    for name in sorted(sys_idx_names):
        sys_idx[name] = len(sys_idx)

    # 1 is added to the indices of all edges in the matrix so that we can use 0 entries to
    # indicate that there is no connection.
    matrix = np.zeros((len(comp_orders), len(comp_orders)), dtype=np.int32)
    edge_ids = []
    for i, edge in enumerate(grp.edges()):
        src, tgt = edge
        if strongdict[src] == strongdict[tgt]:
            matrix[comp_orders[src], comp_orders[tgt]] = i + 1  # bump edge index by 1
            edge_ids.append((sys_idx[src], sys_idx[tgt]))
        else:
            edge_ids.append(None)

    for edge_i, (src, tgt) in enumerate(grp.edges()):
        if strongdict[src] == strongdict[tgt]:
            start = comp_orders[src]
            end = comp_orders[tgt]
            # get a view here, so we can remove this edge from submat temporarily to eliminate
            # an 'if' check inside the nested list comprehension for edges_list
            rem = matrix[start : start + 1, end : end + 1]
            rem[0, 0] = 0

            if end < start:
                start, end = end, start

            submat = matrix[start : end + 1, start : end + 1]
            nz = submat[submat > 0]

            rem[0, 0] = edge_i + 1  # put removed edge back

            if nz.size > 1:
                nz -= 1  # convert back to correct edge index
                edges_list = [edge_ids[i] for i in nz]
                edges_list = sorted(edges_list, key=itemgetter(0, 1))

                for vsrc, vtgtlist in grp.get_edge_data(src, tgt)["conns"].items():
                    for vtgt in vtgtlist:
                        connections_list.append(
                            {"src": vsrc, "tgt": vtgt, "cycle_arrows": edges_list}
                        )
                continue

        for vsrc, vtgtlist in grp.get_edge_data(src, tgt)["conns"].items():
            for vtgt in vtgtlist:
                connections_list.append({"src": vsrc, "tgt": vtgt})

    connections_list = sorted(connections_list, key=itemgetter("src", "tgt"))

    data_dict["sys_pathnames_list"] = list(sys_idx)
    data_dict["connections_list"] = connections_list
    data_dict["abs2prom"] = root_group._var_abs2prom

    data_dict["driver"] = {
        "name": driver_name,
        "type": driver_type,
        "options": driver_options,
        "opt_settings": driver_opt_settings,
    }
    data_dict["design_vars"] = root_group.get_design_vars(use_prom_ivc=False)
    data_dict["responses"] = root_group.get_responses(use_prom_ivc=False)

    data_dict["declare_partials_list"] = _get_declare_partials(root_group)

    return data_dict


def model_to_dict(
    data_source: openmdao.core.problem.Problem,
    path: str or None = None,
    values: bool or _UNDEFINED = _UNDEFINED,
    case_id: str or None = None,
    title: str or None = None,
) -> dict:
    """
    Generate an JSON file containing the model.
    Pulled from N2 Viewer code

    Parameters
    ----------
    data_source : <Problem> or str
        The Problem or case recorder database containing the model or model data.
    path : str, optional
        If specified, the n2 viewer will begin in a state that is zoomed in on the selected path.
        This path should be the absolute path of a system in the model.
    values : bool or _UNDEFINED
        If True, include variable values. If False, all values will be None.
        If unspecified, this behaves as if set to True unless the data source is a Problem or
        model for which setup is not complete, in which case it behaves as if set to False.
    case_id : int, str, or None
        Case name or index of case in SQL file if data_source is a database.
    title : str, optional
        The title for the diagram. Used in the HTML title.

    Returns
    ----------
    json_vars: dict
        Model represented as dict
    """
    # grab the model viewer data
    model_data = _get_viewer_data(data_source, values=values, case_id=case_id)

    options = {}
    model_data["options"] = options

    if title:
        title = f"OpenMDAO Model Hierarchy and N2 diagram: {title}"
    else:
        title = "OpenMDAO Model Hierarchy and N2 diagram"

    json_vars = {
        "title": title,
        "openmdao_version": openmdao_version,
        "model_data": model_data,
        "initial_path": path,
    }

    return json_vars


def write_json(json_vars_: dict, outfile: str) -> None:
    """
    Write model dictionary json_vars_ to file outfile.
    Parameters
    ----------
     json_vars_:  dict
        Dictionary of model description

     param outfile: str
        Fully qualified path of output data file
    Returns
    -------
     Returns nothing

    """
    success_flag = False
    # replace filter_numpy with jason_numpy
    # json_vars_new = filter_numpy(json_vars_)
    json_vars_new = json_vars_
    with open(outfile, "w") as out_dict_f:
        # change numpy values to plain
        out_dict = json.dumps(json_vars_new)
        out_dict_json = json.loads(out_dict)
        out_dict_f.write(json.dumps(out_dict_json, indent=4, sort_keys=True))


def pull_connections_list(full_model_dict: dict) -> list:
    # tested
    """
    Returns a list of the components which make up the system model.

    Parameters
    ----------
    full_model_dict: dict
        Dict of system model

    Returns
    ----------
    circuit_components: list
        List of the components which make up the system model.
    """
    # pull out connections list
    connections_list = full_model_dict["model_data"]["connections_list"]
    # pull system pathnames list to decode connection list cycle arrows
    sys_pathnames_list = full_model_dict["model_data"]["sys_pathnames_list"]

    for d in connections_list:
        if "cycle_arrows" in d:
            temp_list = []
            for prs in d["cycle_arrows"]:
                nams = [
                    sys_pathnames_list[int(prs[0])],
                    sys_pathnames_list[int(prs[1])],
                ]
                temp_list.append(nams)
            d["cycle_names"] = temp_list

    # get all systems
    circuit = full_model_dict["model_data"]["tree"]["children"]
    # get base name
    base_name = list(full_model_dict["model_data"]["responses"].keys())[0].split(".")[0]
    # get component names
    circuit_components = []
    for cir in circuit:
        # if cir["name"].startswith("circuit"):
        if cir["name"].startswith(base_name):
            cir_tok = cir["name"].split(".")
            circuit_components.append(cir_tok[1])
    circuit_components = list(set(circuit_components))

    print("\n")
    for key in circuit_components:
        print(f"Find info in some other file for {key}")
    return circuit_components


def get_additional_options(model_dict: dict, add_option_file_: str) -> dict:
    """
    Reads file containing additional parameters to be added to madel
    TODO: This function needs to be replaced with another that programmatically retrieves the additional parameters.

    Parameters
    ----------
    model_dict: dict
        OpenMDAO system model.

    add_option_file_: str
        Path to file containing a dict of additional options to add to model.

    Returns
    ----------
    """
    # tested
    with open(add_option_file_, "r") as optF:
        additional_options_dict = json.load(optF)
        options_keys = additional_options_dict.keys()
        temp_opt_dict = {}
        for comp in options_keys:
            temp_comp_dict = {}
            for comp_keys in additional_options_dict[comp]:
                temp_comp_dict[comp_keys] = additional_options_dict[comp][comp_keys]
            temp_opt_dict[comp] = temp_comp_dict
        # Add options to model dict
        model_dict["Optional"] = temp_opt_dict
        return model_dict


def get_cycle_names(temp_dict: dict) -> dict:
    """
    Converts cycle_arrow numbers to component connection names
     Parameters
    ----------
     temp_dict:  dict
        Dictionary of model description

    Returns
    -------
     out_dict: dict
        Original dictionary adding cycle_names converted form cycle_arrow numbers
    """
    out_dict = copy.deepcopy(temp_dict)
    # pull out connections list
    connections_list = out_dict["model_data"]["connections_list"]
    # print(f'connections_list:\n{connections_list}')
    ###############################################
    # pull system pathnames list to decode connection list cycle arrows
    sys_pathnames_list = out_dict["model_data"]["sys_pathnames_list"]
    print(f"sys_pathnames_list:\n{sys_pathnames_list}")

    for d in connections_list:
        if "cycle_arrows" in d:
            temp_list = []
            for prs in d["cycle_arrows"]:
                nams = [
                    sys_pathnames_list[int(prs[0])],
                    sys_pathnames_list[int(prs[1])],
                ]
                temp_list.append(nams)
            seen_list = set()
            single_list = [
                x
                for x in temp_list
                if tuple(x[::-1]) not in seen_list and not seen_list.add(tuple(x))
            ]
            d["cycle_names"] = temp_list
            d["cycle_names_single"] = single_list
    return out_dict



"""
Definition of evaluator information Pydantic classes.

Created Dec. 10, 2024

@author Joerg Gablonsky
"""

import re
from typing import List, Optional, Dict, Tuple, Union, Literal
import typing
import numpy as np
from pydantic import BaseModel, Field, field_validator, model_validator
from numpydantic import NDArray, Shape

from standard_evaluator import Variable
from standard_evaluator import unique_names


class EvaluatorInfo(BaseModel):
    """
    Represents information about an evaluator.

    Args:
        name: The name of the evaluator.
        class_type: Identifier
        inputs: Input elements
        outputs: Output elements
        description: A description of the component. To define mathematical symbols use markdown syntax.
        cite: Listing of relevant citations that should be referenced when publishing work that uses this class.
        tool: Name of the tool wrapped
        evaluator_identifier: Unique identifier for the evaluator.
        version: Version of the evaluator.
        component_type: Component type (ExplicitComponent, ImplicitComponent, Group, etc.).
        options: Additional options for the component.
    """

    name: str = Field(description="The name of the evaluator")
    class_type: Literal["EvaluatorInfo"] = "EvaluatorInfo"
    inputs: List[Variable] = Field(description="Input elements")
    outputs: List[Variable] = Field(description="Output elements")
    description: Optional[str] = Field(
        default=None,
        description="A description of the component. To define mathematical symbols use markdown syntax.",
    )
    cite: Optional[str] = Field(
        default=None,
        description="Listing of relevant citations that should be referenced when publishing work that uses this class.",
    )
    tool: Optional[str] = Field(default=None, description="Name of the tool wrapped")
    evaluator_identifier: Optional[str] = Field(
        default=None, description="Unique identifier for the evaluator."
    )
    version: Optional[str] = Field(
        default=None, description="Version of the evaluator."
    )
    component_type: Optional[str] = Field(
        default=None,
        description="Component type (ExplicitComponent, ImplicitComponent, Group, etc.).",
    )
    options: Dict = Field(
        default_factory=dict, description="Additional options for the problem."
    )

    def calculate_default(self, overwrite: bool = True) -> None:
        """Calculate and set default values for all inputs.

        This method iterates through the list of inputs and calls the
        `calculate_default` method on each input to compute its default value.

        Parameters
        ----------
        overwrite : bool
            A flag indicating whether to overwrite existing default values. Defaults to True.
        """
        for input in self.inputs:
            input.calculate_default(overwrite=overwrite)

    def input_names(self) -> List[str]:
        """Retrieve the names of all input inputs.

        This method returns a list of names for all inputs defined in the
        `inputs` attribute.

        Returns
        -------
        List[str]
            A list of names of the input inputs.
        """
        return [element.name for element in self.inputs]
    
    def response_names(self) -> List[str]:
        """Retrieve the names of all outputs.

        This method returns a list of names for all outputs defined in the
        `responses` attribute.

        Returns
        -------
        List[str]
            A list of names of the output outputs.
        """
        return [element.name for element in self.responses]
    
    def set_defaults(self, new_defaults: dict) -> None:
        """Set new default values for input outputs based on a provided dictionary.

        This method updates the default values of the inputs in the `inputs`
        list if their names match the keys in the `new_defaults` dictionary.

        Parameters
        ----------
        new_defaults : dict
            A dictionary where keys are input names and values are the new default values to set.
        """
        for index in range(len(self.inputs)):
            element = self.inputs[index]
            if element.name in new_defaults:
                self.inputs[index].default = new_defaults[element.name]

    @field_validator("inputs", "outputs")
    def validate_outputs(cls, var):
        unique_names(var)
        return var

class EquationInfo(EvaluatorInfo):
    class_type: typing.Literal["EquationInfo"] = "EquationInfo"
    equations : typing.Union[str, typing.List[str]] = Field(description="String or list of strings containing the equation or equations that should be used to calculate the outputs.")


class GroupInfo(EvaluatorInfo):
    class_type: typing.Literal["GroupInfo"] = "GroupInfo"
    component_order: typing.List[str] = Field(description="List of the components in this group in the order they should be displayed")
    components: typing.Dict[str, "JoinedInfo"]
    promotions: typing.Dict[str, typing.List[typing.Tuple[str, str]]] = Field(default = {}, description="Dictionary of lists that map the name of an input / output in a component to an input / output of this group")
    linkage: typing.List[typing.Tuple[str, str]] = Field(default=[], description="Map that allows linkage between inputs and outputs between components inside this group. Needs to use component_name.element_name")

JoinedInfo = typing.Union[ EquationInfo, EvaluatorInfo, GroupInfo]



"""
Definition of optimization problems information Pydantic classes.

Created Nov. 14, 2024

@author Joerg Gablonsky
"""

import re
from typing import List, Optional, Dict, Tuple, Union, Literal, Any
import typing
import numpy as np
from pydantic import (
    BaseModel,
    Field,
    field_validator,
    model_validator,
    field_serializer,
)
from numpydantic import NDArray, Shape
from standard_evaluator import unique_names

MAXINT = 2**63


def generate_names(name: str, shape: tuple) -> typing.List[str]:
    """Generate names for variables / responses that reflect that they are arrays.

    The names generated use NumPy array syntax. For example, a variable input that
    is a 1-d array of length 2 (shape (2,)) will generate this list:

    ['input[0]', 'input[1]']

    Parameters
    ----------
    name : str
        Name of the variable or response.
    shape : tuple
        Shape of the variable or response. This is a NumPy shape.

    Returns
    -------
    list[str]
        List of all the variables / responses represented by the array.

    Raises
    ------
    TypeError
        If the name is not a string, or the shape is not a tuple.
    """
    if not isinstance(name, str):
        raise TypeError("Name is not a string")
    if not isinstance(shape, tuple):
        raise TypeError("Shape is not a tuple")

    new_names = []
    for idx in np.ndindex(shape):
        new_names.append(f"{name}[" + ",".join([str(value) for value in idx]) + "]")
    return new_names


class FloatVariable(BaseModel, validate_assignment=True):
    """Class representing a floating-point variable with bounds and scaling.

    Attributes:
        name: Name of the variable.
        default: Default value for this variable.
        bounds: Lower and upper bounds of the variable.
        shift: Shift value to be used for this variable.
        scale: Scale value to be used for this variable. Cannot be zero.
        units: Units of the variable.
        description: A description of the variable.
        options: Additional options for the variable.
        class_type: Class marker for identifying the variable type.
    """

    name: str = Field(description="Name of the variable.")
    default: Optional[float] = Field(
        default=None, description="Default value for this variable", 
    )
    bounds: Optional[Tuple[float, float]] = Field(
        default=(-np.inf, np.inf), description="Lower and upper bounds of the variable"
    )
    shift: Optional[float] = Field(
        default=0.0,
        description="Shift value to be used for this variable.",
    )
    scale: Optional[float] = Field(
        default=1.0,
        description="Scale value to be used for this variable. Cannot be zero.",
    )
    units: Optional[str] = None
    description: Optional[str] = Field(
        default="",
        description="A description of the variable.",
    )
    options: Dict = Field(default_factory=dict, description="Options for the variable.")
    class_type: Literal["float"] = Field(default="float", description="Class marker")

#    class Config:
#        validate_assignment = True

    def calculate_default(self, overwrite: bool = True) -> None:
        """Calculate the default value based on the bounds.

        If the overwrite flag is not true and default is set, the method exits.
        The default value is set based on the bounds of the variable.

        Parameters
        ----------
        overwrite : bool
            Flag indicating whether to overwrite the default value if it is already set.
        """
        low, high = self.bounds
        if not overwrite and self.default is not None:
            return
        if low == -np.inf:
            if high == np.inf:
                default_value = 0
            else:
                default_value = high
        elif high == np.inf:
            default_value = low
        else:
            default_value = (low + high) / 2
        self.default = default_value

    @field_validator("bounds")
    def validate_bounds(cls, var):
        """Validate the bounds of the variable.

        If the bounds are not set, they default to (-inf, inf). Raises a ValueError
        if the lower bound is greater than the upper bound.

        Parameters
        ----------
        var : Tuple[float, float]
            The bounds to validate.

        Returns
        -------
        Tuple[float, float]
            The validated bounds.

        Raises
        ------
        ValueError
            If the lower bound is greater than the upper bound.
        """
        if var is None:
            return (-np.inf, np.inf)
        if var[0] > var[1]:
            raise ValueError(
                f"Lower bounds are larger than upper bounds: {var[0]} > {var[1]}"
            )
        return var

    @field_validator("scale")
    def validate_scale(cls, scale):
        """Validate the scale of the variable.

        Raises a ValueError if the scale is set to zero.

        Parameters
        ----------
        scale : float
            The scale to validate.

        Returns
        -------
        float
            The validated scale.

        Raises
        ------
        ValueError
            If the scale is zero.
        """
        if scale == 0:
            raise ValueError(f"Scale cannot be 0.")
        return scale

    @field_validator("name")
    def check_name(cls, name: str) -> str:
        """Check the validity of the variable name.

        Raises a ValueError if the name contains whitespace or is an empty string.

        Parameters
        ----------
        name : str
            The name to validate.

        Returns
        -------
        str
            The validated name.

        Raises
        ------
        ValueError
            If the name contains whitespace or is empty.
        """
        if len(name) > 0:
            if re.search(r"\s", name):
                raise ValueError(f"Name cannot contain white spaces.")
        return name


class IntVariable(FloatVariable, validate_assignment=True):
    """Class representing an integer variable with bounds and scaling.

    Inherits from FloatVariable and overrides specific attributes and methods
    for integer handling.
    """

    default: Optional[int] = Field(
        default=None, description="Default value for this variable"
    )
    bounds: Optional[Tuple[int, int]] = Field(
        default=[-MAXINT, MAXINT], description="Lower and upper bounds of the variable"
    )
    shift: Optional[int] = Field(
        default=0,
        description="Shift value to be used for this variable.",
    )
    scale: Optional[int] = Field(
        default=1,
        description="Scale value to be used for this variable. Cannot be zero.",
    )
    class_type: Literal["int"] = Field(default="int", description="Class marker")

    def calculate_default(self, overwrite: bool = True) -> None:
        """Calculate the default value based on the bounds for integer variables.

        If the overwrite flag is not true and default is set, the method exits.
        The default value is set based on the bounds of the variable.

        Parameters
        ----------
        overwrite : bool
            Flag indicating whether to overwrite the default value if it is already set.
        """
        if not overwrite and self.default is not None:
            return
        low, high = self.bounds
        if low == -MAXINT:
            if high == MAXINT:
                default_value = 0
            else:
                default_value = high
        elif high == MAXINT:
            default_value = low
        else:
            default_value = int(np.floor((low + high) / 2))
        self.default = default_value

    @field_validator("bounds")
    def validate_bounds(cls, var):
        """Validate the bounds of the integer variable.

        If the bounds are not set, they default to (-MAXINT, MAXINT). Raises a ValueError
        if the lower bound is greater than the upper bound.

        Parameters
        ----------
        var : Tuple[int, int]
            The bounds to validate.

        Returns
        -------
        Tuple[int, int]
            The validated bounds.

        Raises
        ------
        ValueError
            If the lower bound is greater than the upper bound.
        """
        if var is None:
            return (-MAXINT, MAXINT)
        if var[0] > var[1]:
            raise ValueError(
                f"Lower bounds are larger than upper bounds: {var[0]} > {var[1]}"
            )
        return var


class CategoricalVariable(FloatVariable, validate_assignment=True):
    """Class representing a categorical variable.

    Attributes:
        default: Default value for this variable.
        bounds: Values this variable can take on. Must be defined.
        shift: Shift value to be used for this variable. Not applicable for categorical variables.
        scale: Scale value to be used for this variable. Not applicable for categorical variables.
        class_type: Class marker for identifying the variable type.
    """

    default: Optional[Any] = Field(
        default=None, description="Default value for this variable"
    )
    bounds: List[Any] = Field(
        description="Values this variable or response can take on. Must be defined."
    )
    shift: Literal[None] = Field(
        default=None,
        description="Shift value to be used for this variable. Does not make sense for categorical variables.",
    )
    scale: Literal[None] = Field(
        default=None,
        description="Scale value to be used for this variable. Does not make sense for categorical variables.",
    )
    units: Optional[str] = None
    class_type: Literal["cat"] = Field(default="cat", description="Class marker")

    def calculate_default(self, overwrite: bool = True) -> None:
        """Calculate the default value for categorical variables.

        If the overwrite flag is not true and default is set, the method exits.
        For categorical variables, the first value in bounds is used as the default.

        Parameters
        ----------
        overwrite : bool
            Flag indicating whether to overwrite the default value if it is already set.
        """
        if not overwrite and self.default is not None:
            return
        self.default = self.bounds[0]

    @field_validator("bounds")
    def validate_bounds(cls, var):
        """Validate the bounds of the categorical variable.

        Raises a ValueError if no valid values are defined.

        Parameters
        ----------
        var : List[Any]
            The bounds to validate.

        Returns
        -------
        List[Any]
            The validated bounds.

        Raises
        ------
        ValueError
            If no valid values are defined.
        """
        if len(var) == 0:
            raise ValueError(
                f"Need to define at least one valid value for this variable"
            )
        return var

    @model_validator(mode="after")
    def check_default(self):
        """Check that the default value, if defined, is valid.

        Raises a ValueError if the default value is not in the bounds.

        Returns
        -------
        CategoricalVariable
            The instance of the class after validation.
        """
        bounds = self.bounds
        default = self.default
        if default is not None:
            if default not in bounds:
                raise ValueError(f"Default not in bounds: {default}")
        return self

class ArrayVariable(FloatVariable, validate_assignment=False):
    """Class defining array variables. The underlying data type are NumPy float64 arrays.

    Raises:
        ValueError: Lower bounds greater than upper bounds for at least one element.
        ValueError: Neither shape nor default are set.
        ValueError: Scale for at least one element is set to 0.
        ValueError: Inconsistent shapes for elements in the class.
    """

    shape: Optional[Tuple[int, ...]] = Field(
        default=None, description="Shape of the arrays"
    )
    default: Optional[NDArray[Shape["*,..."], np.float64]] = Field(
        default=None,
        description="NumPy array with the default values to be used for the array variable",
    )
    bounds: Optional[
        Tuple[
            Union[float, int, NDArray[Shape["*,..."], np.float64]],
            Union[float, int, NDArray[Shape["*,..."], np.float64]],
        ]
    ] = Field(default=(None, None), description="Lower and upper bounds of the arrays")
    shift: Optional[Union[float, int, NDArray[Shape["*,..."], np.float64]]] = Field(
        default=None,
        description="NumPy array with the shift values to be used for the array variable.",
    )
    scale: Optional[Union[float, int, NDArray[Shape["*,..."], np.float64]]] = Field(
        default=None,
        description="NumPy array with the scale values to be used for the array variable. "
        + "Cannot be zero.",
    )
    class_type: Literal["floatarray"] = "floatarray"

    def calculate_default(self, overwrite: bool = True) -> None:
        """Calculate the default value based on the bounds for array variables.

        If the overwrite flag is not true and default is set, the method exits.
        The default value is calculated based on the bounds of the variable.

        Parameters
        ----------
        overwrite : bool
            Flag indicating whether to overwrite the default value if it is already set.
        """
        if not overwrite and self.default is not None:
            return
        low, high = self.bounds
        # Create a new array for the average
        default_value = np.where(
            (low == -np.inf) & (high == np.inf), 0,
            np.where(
                low == -np.inf, high,
                np.where(
                    high == np.inf, low,
                    (low + high) / 2
                )
            )
        )
        self.default = default_value
        
    @field_validator("bounds")
    def validate_bounds(cls, var):
        """Validate the bounds of the array variable.

        If the bounds are not set, they default to (None, None). Raises a ValueError
        if any lower bound is greater than the corresponding upper bound.

        Parameters
        ----------
        var : Tuple[Union[float, int, NDArray[Shape["*,..."], np.float64]], Union[float, int, NDArray[Shape["*,..."], np.float64]]]
            The bounds to validate.

        Returns
        -------
        Tuple[Union[float, int, NDArray[Shape["*,..."], np.float64]], Union[float, int, NDArray[Shape["*,..."], np.float64]]]
            The validated bounds.

        Raises
        ------
        ValueError
            If any lower bound is greater than the corresponding upper bound.
        """
        if var is None:
            return (None, None)
        if np.any(var[0] > var[1]):
            raise ValueError(
                f"Lower bounds are larger than upper bounds: {var[0]} > {var[1]}"
            )
        return var

    @field_validator("scale")
    def validate_scale(cls, scale):
        """Validate the scale of the array variable.

        This method does not perform any checks as validation occurs in the check_shape method.

        Parameters
        ----------
        scale : Union[float, int, NDArray[Shape["*,..."], np.float64]]
            The scale to validate.

        Returns
        -------
        Union[float, int, NDArray[Shape["*,..."], np.float64]]
            The validated scale.
        """
        return scale

    def _expand(self, value, scaler=1.0):
        """Expand a single value to match the shape of the array.

        If the value is None, it is replaced with an array of the specified shape filled with the scaler value.

        Parameters
        ----------
        value : Union[float, int, None]
            The value to expand.
        scaler : float, optional
            The value to fill the array if value is None. Defaults to 1.0.

        Returns
        -------
        NDArray[Shape["*,..."], np.float64]
            The expanded array.
        """
        if value is not None:
            if isinstance(value, (float, int)):
                value = np.ones(self.shape) * value
        else:
            value = np.ones(self.shape) * scaler
        return value

    @model_validator(mode="after")
    def check_shape(self):
        """Check and validate the shape of the array variable.

        This method ensures that the shape is defined and consistent across default, shift, scale, and bounds.

        Returns
        -------
        ArrayVariable
            The instance of the class after validation.

        Raises
        ------
        ValueError
            If the shape is not set and neither default nor value is defined.
            If any of the shapes for default, shift, scale, or bounds are inconsistent.
        """
        shape = self.shape
        default = self.default
        shift = self.shift
        scale = self.scale

        if shape is None:
            # If no shape is defined  default must be set. If default is
            # set it's shape will define the shape. If neither is set we throw an error
            if default is not None:
                shape = default.shape
            else:
                raise ValueError(
                    f"Shape for this variable is not set, and neither are default or value"
                )
            self.shape = shape

        # If shift is set we want to check if only a single value is set. If that is the
        # case we expand it to the full shape.
        shift = self._expand(shift, 0.0)
        self.shift = shift

        # If scale is set we want to check if only a single value is set. If that is the
        # case we expand it to the full shape.
        scale = self._expand(scale)

        # We have to make sure none of the values of scale are 0.
        if np.any(scale == 0.0):
            raise ValueError(f"Scale for one element in the matrix is set to 0.0")

        self.scale = scale

        # Check that the bounds are set correctly
        (low_bound, up_bound) = self.bounds
        low_bound = self._expand(low_bound, -np.inf)
        up_bound = self._expand(up_bound, np.inf)
        self.bounds = (low_bound, up_bound)

        fields = [default, shift, scale, low_bound, up_bound]
        field_names = ["default", "shift", "scale", "lower bound", "upper bound"]
        # Check if default is not None. If default is set we check that the shape is
        # set correctly.
        for local_field, field_name in zip(fields, field_names):
            if local_field is not None:
                if local_field.shape != shape:
                    raise ValueError(
                        f"Shape for this variable is {shape}, but {field_name} has shape {local_field.shape}"
                    )

        return self

    @field_serializer("bounds")
    def serialize_bounds(
        self,
        bounds: Tuple[
            NDArray[Shape["*,..."], np.float64],
            NDArray[Shape["*,..."], np.float64],
        ],
    ):
        """Serialize the bounds for storage.

        If the bounds are set to (-inf, inf), they are not stored. Otherwise, the bounds are
        serialized to a more compact form if possible.

        Parameters
        ----------
        bounds : Tuple[NDArray[Shape["*,..."], np.float64], NDArray[Shape["*,..."], np.float64]]
            The bounds to serialize.

        Returns
        -------
        Optional[Tuple[Union[float, int], Union[float, int]]]
            The serialized bounds, or None if they are set to (-inf, inf).
        """
        if np.all(np.isinf(bounds[0])) & np.all(np.isinf(bounds[0])):
            return None
        else:
            lower = bounds[0]
            upper = bounds[1]
            if np.all(upper == upper.ravel()[0]):
                upper = upper.ravel()[0]
            if np.all(lower == lower.ravel()[0]):
                lower = lower.ravel()[0]
            bounds = (lower, upper)
            return bounds

    @field_serializer("shift")
    def serialize_shift(self, shift: NDArray[Shape["*,..."], np.float64]):
        """Serialize the shift values for storage.

        If the shift is set to all zeros, it is not stored. Otherwise, the shift values are returned.

        Parameters
        ----------
        shift : NDArray[Shape["*,..."], np.float64]
            The shift values to serialize.

        Returns
        -------
        Optional[NDArray[Shape["*,..."], np.float64]]
            The serialized shift values, or None if they are all zeros.
        """
        if np.all(shift == 0.0):
            return None
        else:
            return shift

    @field_serializer("scale")
    def serialize_scale(self, scale: NDArray[Shape["*,..."], np.float64]):
        """Serialize the scale values for storage.

        If the scale is set to all ones, it is not stored. Otherwise, the scale values are returned.

        Parameters
        ----------
        scale : NDArray[Shape["*,..."], np.float64]
            The scale values to serialize.

        Returns
        -------
        Optional[NDArray[Shape["*,..."], np.float64]]
            The serialized scale values, or None if they are all ones.
        """
        if np.all(scale == 1.0):
            return None
        else:
            return scale

# Define the Union of the different variable types. Note that we use that for responses as well
Variable = Union[FloatVariable, IntVariable, ArrayVariable, CategoricalVariable]


from pydantic import BaseModel, Field
from typing import List, Dict, Optional, Literal

class OptProblem(BaseModel):
    """
    Represents information about an optimization problem.

    Args:
        name: The name of the problem. Defaults to 'opt_problem'.
        variables: A list of input variables.
        responses: A list of output variables.
        objectives: Names of the objective(s) for the optimization problem. Must be either variables or responses defined in the problem.
        constraints: Names of the constraints of the optimization problem. Must be responses defined in the problem. To define bounds on variables use the variable bounds.
        description: A description of the optimization problem. To define mathematical symbols use markdown syntax.
        cite: Listing of relevant citations that should be referenced when publishing work that uses this class.
        options: Additional options for the problem.
    """

    name: str = Field(
        default="opt_problem",
        description='The name of the problem. Defaults to "opt_problem".',
    )
    class_type: Literal["OptProblem"] = "OptProblem"
    variables: List[Variable] = Field(description="Input variables")
    responses: List[Variable] = Field(description="Output variables")
    objectives: List[str] = Field(
        default_factory=list,
        description="Names of the objective(s) for the optimization problem. Must be either variables or responses defined in the problem.",
    )
    constraints: List[str] = Field(
        default_factory=list,
        description="Names of the constraints of the optimization problem. Must be responses defined in the problem. To define bounds on variables use the variable bounds.",
    )
    description: Optional[str] = Field(
        default=None,
        description="A description of the optimization problem. To define mathematical symbols use markdown syntax.",
    )
    cite: Optional[str] = Field(
        default=None,
        description="Listing of relevant citations that should be referenced when publishing work that uses this class.",
    )
    options: Dict = Field(
        default_factory=dict, description="Additional options for the problem."
    )

    def calculate_default(self, overwrite: bool = True) -> None:
        """Calculate and set default values for all input variables.

        This method iterates through the list of variables and calls the
        `calculate_default` method on each variable to compute its default value.

        Parameters
        ----------
        overwrite : bool
            A flag indicating whether to overwrite existing default values. Defaults to True.
        """
        for variable in self.variables:
            variable.calculate_default(overwrite=overwrite)

    def variable_names(self) -> List[str]:
        """Retrieve the names of all input variables.

        This method returns a list of names for all variables defined in the
        `variables` attribute.

        Returns
        -------
        List[str]
            A list of names of the input variables.
        """
        return [element.name for element in self.variables]
    
    def response_names(self) -> List[str]:
        """Retrieve the names of all output variables.

        This method returns a list of names for all variables defined in the
        `responses` attribute.

        Returns
        -------
        List[str]
            A list of names of the output variables.
        """
        return [element.name for element in self.responses]
    
    def set_defaults(self, new_defaults: dict) -> None:
        """Set new default values for input variables based on a provided dictionary.

        This method updates the default values of the variables in the `variables`
        list if their names match the keys in the `new_defaults` dictionary.

        Parameters
        ----------
        new_defaults : dict
            A dictionary where keys are variable names and values are the new default values to set.
        """
        for index in range(len(self.variables)):
            element = self.variables[index]
            if element.name in new_defaults:
                self.variables[index].default = new_defaults[element.name]


    @field_validator("variables", "responses")
    def validate_outputs(cls, var):
        """Validate the names of the variables and responses.

        This method checks that the names of the variables and responses are unique
        using the `unique_names` function.

        Parameters
        ----------
        var : List[Variable]
            The list of variables or responses to validate.

        Returns
        -------
        List[Variable]
            The validated list of variables or responses.
        """
        unique_names(var)
        return var

    def unroll_names(self, elements: List[Variable]) -> List[str]:
        """Unroll the names of the variables and responses into a flat list.

        This method generates a list of names for all elements, including
        expanding array variables into their individual components.

        Parameters
        ----------
        elements : List[Variable]
            The list of variables or responses to unroll.

        Returns
        -------
        List[str]
            A flat list of names for the variables and responses.
        """
        all_names = []
        for local_element in elements:
            name = local_element.name
            if isinstance(local_element, ArrayVariable):
                array_names = generate_names(name, local_element.shape)
                all_names += array_names
            else:
                all_names.append(name)
        return all_names

    @model_validator(mode="after")
    def check_problem(self):
        """Check the validity of the optimization problem setup.

        This method verifies that all objectives and constraints are defined correctly.
        It checks that objectives are either variables or responses and that constraints
        are defined as responses.

        Returns
        -------
        OptProblem
            The instance of the class after validation.

        Raises
        ------
        ValueError
            If any objective is not defined as a variable or response.
            If any constraint is defined as a variable instead of a response.
        """
        variable_names = self.unroll_names(self.variables)
        response_names = self.unroll_names(self.responses)
        elements = set(variable_names + response_names)

        if self.objectives is not None:
            # Check if all the objectives are either a variable or response
            for name in self.objectives:
                if name not in elements:
                    raise ValueError(
                        f"{name} is defined as an objective, but not defined as a variable or response."
                    )

        if self.constraints is not None:
            # Check if all the objectives are either a variable or response
            for name in self.constraints:
                if name not in response_names:
                    if name in variable_names:
                        raise ValueError(
                            f"{name} is defined as a constraint, but defined as a variable. Please define bounds on the variable itself. Constraints should only be responses."
                        )
                    raise ValueError(
                        f"{name} is defined as a constraint, but not defined as a response."
                    )

        return self


""" Implementing the Standard Evaluator for components    """
import inspect
import typing
from typing import List

from openmdao.core.constants import _UNDEFINED
from openmdao.visualization.tables.table_builder import generate_table
from openmdao.utils.options_dictionary import OptionsDictionary

TAG_KEY = "Options:"

class OptionsDictionaryUnit(OptionsDictionary):
    """Expand the OptionsDictionary to allow the definition of Units

Arguments:
    OptionsDictionary {[type]} -- [description]

Raises:
    KeyError: [description]
    TypeError: [description]
    TypeError: [description]
    ValueError: [description]
    ValueError: [description]

Returns:
    [type] -- [description]
    """
    def declare(self, name, default=_UNDEFINED, values=None, types=None, desc='',
                units="unitless", 
                upper=None, lower=None, allow_none=False, recordable=True,
                deprecation=None):
        r"""
        Declare an option.

        The value of the option must satisfy the following:
        1. If values only was given when declaring, value must be in values.
        2. If types only was given when declaring, value must satisfy isinstance(value, types).
        3. It is an error if both values and types are given.

        Parameters
        ----------
        name : str
            Name of the option.
        default : object or Null
            Optional default value that must be valid under the above 3 conditions.
        values : set or list or tuple or None
            Optional list of acceptable option values.
        types : type or tuple of types or None
            Optional type or list of acceptable option types.
        desc : str
            Optional description of the option.
        units : str
            Optional definition of units. Defaults to `unitless`
        upper : float or None
            Maximum allowable value.
        lower : float or None
            Minimum allowable value.
        check_valid : function or None
            User-supplied function with arguments (name, value) that raises an exception
            if the value is not valid.
        allow_none : bool
            If True, allow None as a value regardless of values or types.
        recordable : bool
            If True, add to recorder.
        set_function : None or function
            User-supplied function with arguments (Options metadata, value) that pre-processes
            value and returns a new value.
        deprecation : str or None
            If None, it is not deprecated. If a str, use as a DeprecationWarning
            during __setitem__ and __getitem__.  If a tuple of the form (msg, new_name),
            display msg as with str, and forward any __setitem__/__getitem__ to new_name.
        """
        if values is not None and not isinstance(values, (set, list, tuple)):
            self._raise(f"In declaration of option '{name}', the 'values' arg must be of type None,"
                        f" list, or tuple - not {values}.", exc_type=TypeError)

        if types is not None and not isinstance(types, (type, set, list, tuple)):
            self._raise(f"In declaration of option '{name}', the 'types' arg must be None, a type "
                        f"or a tuple - not {types}.", exc_type=TypeError)

        if types is not None and types is not list and values is not None:
            self._raise(f"'types' and 'values' were both specified for option '{name}'.")

        if types is bool:
            values = (True, False)

        if not recordable:
            self._all_recordable = False

        default_provided = default is not _UNDEFINED

        if default_provided and default is None:
            # specifying default=None implies allow_none
            allow_none = True

        alias = None
        if deprecation is not None:
            if isinstance(deprecation, (list, tuple)):
                if len(deprecation) != 2:
                    self._raise("deprecation must be None, str, or a tuple or list containing "
                                "(str, str).", RuntimeError)
                dep, alias = deprecation
                # [message, alias, display warning (becomes False after first display)]
                deprecation = [dep, alias, True]
            else:
                deprecation = [deprecation, None, True]

        self._dict[name] = {
            'val': default,
            'values': values,
            'types': types,
            'desc': desc,
            'upper': upper,
            'lower': lower,
            'check_valid': None,
            'has_been_set': default_provided,
            'allow_none': allow_none,
            'recordable': recordable,
            'set_function': None,
            'deprecation': deprecation,
            'units': units
        }

        # If a default is given, check for validity
        if default_provided:
            self._assert_valid(name, default)

    def get_item(self, name):
        value = self[name]
        units = self._dict[name]["units"]
        return(value, units)

    def to_table(self, fmt='github', missingval='N/A', max_width=None, display=True):
        """
        Get a table representation of this OptionsDictionary as a table in the requested format.

        Parameters
        ----------
        fmt : str
            The formatting of the requested table.  Options are
            ['github', 'rst', 'text', 'html', 'tabulator'] and several 'grid' and 'outline'
            formats that mimic those found in the python 'tabulate' library.
            Default value of 'github' produces a table in GitHub-flavored markdown.
            'html' and 'tabulator' produce output viewable in a browser.
        missingval : str
            The value to be displayed in place of None.
        max_width : int or None
            If not None, try to limit the total width of the table to this value.
        display : bool
            If True, display the table, typically by writing it to stdout or opening a
            browser.

        Returns
        -------
        str
            A string representation of the table in the requested format.
        """
        hdrs = ['Option', 'Default', 'Acceptable Values', 'Acceptable Types', 'Description', 'Units']
        rows = []

        deprecations = False
        for meta in self._dict.values():
            if meta['deprecation'] is not None:
                deprecations = True
                hdrs.append('Deprecation')
                break

        for key in sorted(self._dict.keys()):
            option = self._dict[key]
            default = option['val'] if option['val'] is not _UNDEFINED else '**Required**'
            default_str = str(default)

            # if the default is an object instance, replace with the (unqualified) object type
            idx = default_str.find(' object at ')
            if idx >= 0 and default_str[0] == '<':
                parts = default_str[:idx].split('.')
                default = parts[-1]

            acceptable_values = option['values']
            if acceptable_values is not None:
                if not isinstance(acceptable_values, (set, tuple, list)):
                    acceptable_values = (acceptable_values,)
                acceptable_values = [value for value in acceptable_values]

            acceptable_types = option['types']
            if acceptable_types is not None:
                if not isinstance(acceptable_types, (set, tuple, list)):
                    acceptable_types = (acceptable_types,)
                acceptable_types = [type_.__name__ for type_ in acceptable_types]

            desc = option['desc']

            deprecation = option['deprecation']
            if deprecation is not None:
                deprecation = deprecation[0]
            units = option['units']

            if deprecations:
                rows.append([key, default, acceptable_values, acceptable_types, desc, units,
                             deprecation])
            else:
                rows.append([key, default, acceptable_values, acceptable_types, desc, units])
        
        if len(rows) == 0:
            # This is an empty class
            rows.append(['']*len(hdrs))

        kwargs = {
            'tablefmt': fmt,
            'headers': hdrs,
            'missing_val': missingval,
            'max_width': max_width,
        }
        if fmt == 'tabulator':
            kwargs['filter'] = False
            kwargs['sort'] = False

        tab = generate_table(rows, **kwargs)

        if display:
            tab.display()

        return str(tab)


def opt_dict_unit_merge(options : List[OptionsDictionaryUnit], 
    base_options : typing.Optional[OptionsDictionaryUnit] = None,
    ) -> OptionsDictionaryUnit:
    """[summary]

    Arguments:
        options {List[OptionsDictionaryUnit]} -- A list of options to merge. Note that if options
                            are defined more than once the final values are used.

    Keyword Arguments:
        base_options {typing.Optional[OptionsDictionaryUnit]} -- If used these options are expanded
                            with the passed in options. (default: {None})
    Returns:
        OptionsDictionaryUnit -- Merge of all the options passed in.
    """
    if base_options is None:
        # We create an empty OptionsDictionaryUnit object to collect all options
        base_options = OptionsDictionaryUnit()
    for single_option in options:
        # Iterating over all options passed in
        for (name, info) in single_option._dict.items():
            # Declaring all options that have been defined in this single_option
            values = info["values"]
            if info['types']  is not None:
                if issubclass(info['types'], bool):
                    values = None
            base_options.declare(name=name, default=info["val"], values=values,
            types=info["types"], desc=info["desc"], upper=info["upper"],
            lower=info["lower"], allow_none=info["allow_none"], recordable=info["recordable"],
            deprecation=info["deprecation"])
    return(base_options)



class StandardBase(object):
    """A class to be used in conjunction with OpenMDAO elements to have a common way of dealing with options.

    This class implements the ability of the system to keep track of options,
    and also pre-defines that this component depends on the 'class_options'.
    """

    def initialize(self):
        """Initialize the instance by defining that this will use `class_options`"""
        self.options.declare(
            "class_options",
            types=OptionsDictionaryUnit,
            desc="collection of Aircraft/Mission specific options",
        )
        # Set the options based on the class method  to define the options.
        self.options["class_options"] = self.required_options()

    @classmethod
    def _define_options(cls) -> OptionsDictionaryUnit:
        """Abstract method that allows a developer to define information about the parameters this class uses.

        This method is called by a class method that is adding options to the `class_options` option.

        Returns:
            OptionsDictionaryUnit -- The required options for this class, their default values, and, if required, their units.
        """
        return OptionsDictionaryUnit()

    @classmethod
    def required_options(cls) -> OptionsDictionaryUnit:
        """A class method that will return the required options and their default values for
        this class.

        Returns:
            OptionsDictionaryUnit: Information about the required options and their default values
        """
        return cls._define_options()

    @classmethod
    def required_options_names(cls) -> set[str]:
        """A class method returning a set of strings with the names of the required options.

        Returns:
            set(str): Names of the required options.
        """
        return set(cls.required_options()._dict.keys())


    def full_options(self) -> OptionsDictionaryUnit:
        """An instance method to return all options for this class.
        
        Often this will be just a call to the `required_options` class method. But for example
        in group the full set of options that are valid are depending on the components in the
        instance, so this has to be calculated after an instance is created, and the `setup()`
        method has been called.

        Returns:
            OptionsDictionaryUnit -- All valid options for this class.
        """
        return self.required_options()

    @property
    def full_options_names(self) -> set[str]:
        """A class method returning a set of strings with the names of the full options.

        Returns:
            set(str): Names of the required options.
        """
        return set(self.full_options()._dict.keys())

    def lookup_option_value(self, name) -> typing.Any:
        class_options: OptionsDictionaryUnit = self.options["class_options"]
        if name in class_options:
            value = class_options[name]
        else:
            # Let's see if the value is set in the default full options
            required_opts = self.full_options()
            if name in required_opts:
                print(f"Looking up value for {name} from the full options method")
                value = required_opts[name]
            else:
                raise KeyError(f'KeyError: key not found: {name}')
        return value

    
    def _check_options(
        self, options: typing.Optional[typing.Union[str, list[str]]]
    ) -> list[str]:
        """Check that the options are correct, and part of the required options.

        Arguments:
            options {typing.Optional[typing.Union[str, list[str]]]} -- Name or list of
                        names for options. Defaults to None.

        Raises:
            TypeError: Options need to be a string or list of strings
            TypeError: Entry in the options list is not a string
            ValueError: Trying to define an option for a variable or response that is not part of the
                required options list. Options need to be defined.

        Returns:
            list[str] -- List of options.
        """
        # Get the names of required options given.
        req_options = self.required_options_names()
        if isinstance(options, str):
            # If only a single string is passed in we convert it into a list of strings
            options = [options]
        elif not isinstance(options, list):
            raise TypeError("Options is not a string or list of strings")
        # Check if the option(s) defined for this variable or response are are part of the
        # named required options. If not throw an error. This ensures that the required
        # options set contains all required options.
        print(req_options)
        for entry in options:
            if not isinstance(entry, str):
                raise TypeError(
                    f"Entry {entry} in options list is not a string, instead a {type(entry)}"
                )
            if not entry in req_options:
                raise ValueError(
                    f"Trying to set {entry} as an option, but it is not defined as an option for this class"
                )
        return options

    def _lookup_info(
        self,
        name: str,
        val: typing.Optional[typing.Any] = None,
        units: typing.Optional[str] = None,
        desc: typing.Optional[str] = None,
    ) -> typing.Tuple[typing.Any, str, str]:
        """Helper method to allow implementation of a way to look up information.

        Arguments:
            name {str} -- Name of the element to look up information for

        Keyword Arguments:
            val {typing.Optional[typing.Any]} -- Value to use. If None value will be looked up (default: {None})
            units {typing.Optional[str]} --Units to use. If None units will be looked up (default: {None})
            desc {typing.Optional[str]} -- Description to use. If None description will be looked up (default: {None})

        Returns:
            typing.Tuple[typing.Any, str, str] -- Updated values, units, and description. 
        """
        # In this base class implementation we are not doing any lookup
        return (val, units, desc)

    def _add_info(
        self,
        name: str,
        element_type: typing.Optional[str] = None,
        val: typing.Optional[typing.Any] = None,
        units: typing.Optional[str] = None,
        desc: typing.Optional[str] = None,
        shape_by_conn: bool = False,
        options: typing.Optional[typing.Union[str, list[str]]] = None,
        look_up: bool = True,
    ):
        """Provide a clean way to add variables or responses from the  hierarchy into
        components as inputs, and allow the definition of options this
        variable or response depends on. It takes
        the standard OpenMDAO inputs of the variable's or response's name, initial
        value, units, and description, as well as the component which
        the variable or response is being added to.

        Args:
            name (str): OpenMDAO variable or response name
            param_type (typing.Optional[str], optional): Description if information is for a
                    variable or response. Defaults to None.
            val (typing.Optional[typing.Any], optional): Initial value. If none the initial value might
                        be provided by the _lookup_info method. Defaults to None.
            units (typing.Optional[str], optional): Units. If none the initial value might
                        be provided by the _lookup_info method. Defaults to None.
            desc (typing.Optional[str], optional): Description. If none the initial value might
                        be provided by the _lookup_info method. Defaults to None.
            shape_by_conn (bool, optional): Decide whether or not the shape of the variable or
                        response is defined by its connections. Defaults to False.
            options (typing.Optional[typing.Union[str, list[str]]], optional): Name or list of
                        names for options this variable or response depends on. Defaults to
                        None.
            look_up (bool, optional): Flag whether to look up values, units, and description. Defaults
                        to true
        """
        if look_up:
            # Use either the passed in values, or the values from the look-up method
            (val, units, desc) = self._lookup_info(
                name=name,
                val=val,
                units=units,
                desc=desc,
            )
        if element_type not in ["variables", "responses"]:
            raise ValueError(
                f"Option type is {element_type}, only `variables` or `responses is allowed."
            )
        if element_type == "variables":
            func_name = self.add_input
        else:
            func_name = self.add_output

        tags = None
        if options is not None:
            # Check that the options are valid list of required options
            options = self._check_options(options)
            # Save all the options in tags, using the TAG_KEY for uniqueness.
            tags = [f"{TAG_KEY}{opt}" for opt in options]
        # Call either the add_input or add_output routine that is defined by the OpenMDAO
        # base object
        func_name(
            name,
            val=val,
            units=units,
            desc=desc,
            shape_by_conn=shape_by_conn,
            tags=tags,
        )

    def add_standard_input(
        self,
        name: str,
        val: typing.Optional[typing.Any] = None,
        units: typing.Optional[str] = None,
        desc: typing.Optional[str] = None,
        shape_by_conn: bool = False,
        options: typing.Optional[typing.Union[str, list[str]]] = None,
        look_up: bool = True,
    ):
        """Provide a clean way to add variables from the variable hierarchy into
        components as Aviary inputs, and allow the definition of options this
        variable depends on. It takes
        the standard OpenMDAO inputs of the variable's name, initial
        value, units, and description, as well as the component which
        the variable is being added to.

        Args:
            name (str): OpenMDAO variable name
            val (typing.Optional[typing.Any], optional): Initial value of the variable. If none
                        the initial value will be taken from the MetaData. Defaults to None.
            units (typing.Optional[str], optional): Units for the variable. If none the units will
                        be taken from the MetaData. Defaults to None.
            desc (typing.Optional[str], optional): Description of the variable. If none the
                        description will be taken from the MetaData. Defaults to None.
            shape_by_conn (bool, optional): Decide whether or not the shape of the variable is
                        defined by its connections. Defaults to False.
            options (typing.Optional[typing.Union[str, list[str]]], optional): Name or list of
                        names for options this variable depends on. Defaults to None.
            look_up (bool, optional): Flag whether to look up values, units, and description. Defaults
                        to true
        """
        # Use either the passed in values, or the values from the metadata
        self._add_info(
            name=name,
            element_type="variables",
            val=val,
            units=units,
            desc=desc,
            shape_by_conn=shape_by_conn,
            options=options,
            look_up=look_up,
        )

    def add_standard_output(
        self,
        name: str,
        val: typing.Optional[typing.Any] = None,
        units: typing.Optional[str] = None,
        desc: typing.Optional[str] = None,
        shape_by_conn: bool = False,
        options: typing.Optional[typing.Union[str, list[str]]] = None,
        look_up: bool = True,
    ):
        """Provide a clean way to add responses from the response hierarchy into
        components as Aviary outputs, and allow the definition of options this
        response depends on. It takes
        the standard OpenMDAO inputs of the response's name, initial
        value, units, and description, as well as the component which
        the response is being added to.

        Args:
            name (str): OpenMDAO response name
            val (typing.Optional[typing.Any], optional): Initial value of the response. If none
                        the initial value will be taken from the MetaData. Defaults to None.
            units (typing.Optional[str], optional): Units for the response. If none the units will
                        be taken from the MetaData. Defaults to None.
            desc (typing.Optional[str], optional): Description of the response. If none the
                        description will be taken from the MetaData. Defaults to None.
            shape_by_conn (bool, optional): Decide whether or not the shape of the response is
                        defined by its connections. Defaults to False.
            options (typing.Optional[typing.Union[str, list[str]]], optional): Name or list of
                        names for options this response depends on. Defaults to None.
            look_up (bool, optional): Flag whether to look up values, units, and description. Defaults
                        to true
        """
        # Use either the passed in values, or the values from the metadata
        self._add_info(
            name=name,
            element_type="responses",
            val=val,
            units=units,
            desc=desc,
            shape_by_conn=shape_by_conn,
            options=options,
            look_up=look_up,
        )

    def _get_options_info(self) -> dict:
        # Create a dictionary that collects the information for all options
        option_names = self.full_options_names
        defined_names = [info for (info,_) in self.options["class_options"].items()]
        option_info = {}
        for name in option_names:
            # Get the values for the options as they are at this point. This will ensure wer get
            # the actual values of the options as they were when the class is fully set.
            if name in defined_names:
                # This option is still set in the "class_options". This means it probably was
                # overwritten by the user, or the user created the initial options from the
                # required_options class method
                (val, units) = self.options["class_options"].get_item(name)
            else:
                # This option was not set, and is not actually used as an option.
                # We go back to the values and units defined in the required_options class method.
                (val, units) = self.full_options().get_item(name)
            option_info[name] = {
                "input": set(),
                "output": set(),
                "value": val,
                "units": units,
            }
        # Iterate over all elements that are in the object. Do this for inputs and outputs
        for element_type in ["input", "output"]:
            for name, local_info in self.get_io_metadata(iotypes=element_type).items():
                # If there are actual entries in the tags we check them
                if len(local_info["tags"]) > 0:
                    # This element has tags associated with it
                    for option in option_names:
                        # Check if this option is used for this element
                        if f"{TAG_KEY}{option}" in local_info["tags"]:
                            option_info[option][element_type].add(name)

        # Need to convert any sets to a list to be "jsonifiable". Also remove empty sets
        for option in option_names:
            for element_type in ["input", "output"]:
                if len(option_info[option][element_type]) > 0:
                    option_info[option][element_type] = list(
                        option_info[option][element_type]
                    )
                else:
                    del option_info[option][element_type]
        return option_info
    
    def current_options(self) -> OptionsDictionaryUnit:
        """Return all the current options. 
        This is a combination between the original options (defaults) and values that
        were overwritten by the end user.

        Returns:
            OptionsDictionaryUnit -- The combination of all options
        """
        return(opt_dict_unit_merge([self.full_options(), self.options["class_options"]]))
